import argparse
import concurrent.futures
import os
import time

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.mlab import detrend
from matplotlib.mlab import detrend_none
from matplotlib.mlab import window_hanning
from scipy.fft import fft  # pylint disable=no-name-in-module
from scipy.fft import fftfreq  # pylint disable=no-name-in-module

from gamutrf.utils import get_nondot_files
from gamutrf.utils import is_fft
from gamutrf.utils import replace_ext
from gamutrf.sample_reader import read_recording, parse_filename


def stride_windows(x, n, noverlap=0, axis=0):
    # np>=1.20 provides sliding_window_view, and we only ever use axis=0.
    if hasattr(np.lib.stride_tricks, "sliding_window_view") and axis == 0:
        if noverlap >= n:
            raise ValueError("noverlap must be less than n")
        return np.lib.stride_tricks.sliding_window_view(x, n, axis=0)[:: n - noverlap].T

    if noverlap >= n:
        raise ValueError("noverlap must be less than n")
    if n < 1:
        raise ValueError("n cannot be less than 1")

    x = np.asarray(x)

    if n == 1 and noverlap == 0:
        if axis == 0:
            return x[np.newaxis]
        else:
            return x[np.newaxis].T
    if n > x.size:
        raise ValueError("n cannot be greater than the length of x")

    # np.lib.stride_tricks.as_strided easily leads to memory corruption for
    # non integer shape and strides, i.e. noverlap or n. See #3845.
    noverlap = int(noverlap)
    n = int(n)

    step = n - noverlap
    if axis == 0:
        shape = (n, (x.shape[-1] - noverlap) // step)
        strides = (x.strides[0], step * x.strides[0])
    else:
        shape = ((x.shape[-1] - noverlap) // step, n)
        strides = (step * x.strides[0], x.strides[0])
    return np.lib.stride_tricks.as_strided(x, shape=shape, strides=strides)


def spectral_helper(
    x,
    NFFT=None,
    Fs=None,
    detrend_func=None,
    window=None,
    noverlap=None,
    pad_to=None,
    scale_by_freq=None,
    mode=None,
    skip_fft=False,
):
    if Fs is None:
        Fs = 2
    if noverlap is None:
        noverlap = 0
    if detrend_func is None:
        detrend_func = detrend_none
    if window is None:
        window = window_hanning

    window = np.abs(window(np.ones(NFFT, np.complex64)))
    if len(window) != NFFT:
        raise ValueError("The window length must match the data's first dimension")

    # if NFFT is set to None use the whole signal
    if NFFT is None:
        NFFT = 256

    if mode is None or mode == "default":
        mode = "psd"

    if pad_to is None:
        pad_to = NFFT

    if mode != "psd":
        scale_by_freq = False
    elif scale_by_freq is None:
        scale_by_freq = True

    # For real x, ignore the negative frequencies unless told otherwise
    numFreqs = pad_to
    if pad_to % 2:
        freqcenter = (pad_to - 1) // 2 + 1
    else:
        freqcenter = pad_to // 2

    results = []

    for i in x:
        if i is None:
            break
        if skip_fft:
            # TODO: assume NFFT is factor of sps.
            result = i.reshape(-1, int(len(i) / NFFT), order="F")
        else:
            result = stride_windows(i, NFFT, noverlap, axis=0)
            result = detrend(result, detrend_func, axis=0)
            result = fft(result, n=pad_to, axis=0)[
                :numFreqs, :
            ]  # pylint: disable=invalid-sequence-index

        if mode == "psd":
            result *= np.conj(result)
            # MATLAB divides by the sampling frequency so that density function
            # has units of dB/Hz and can be integrated by the plotted frequency
            # values. Perform the same scaling here.
            if scale_by_freq:
                result /= Fs
                # Scale the spectrum by the norm of the window to compensate for
                # windowing loss; see Bendat & Piersol Sec 11.5.2.
                result /= (window**2).sum()
            else:
                # In this case, preserve power in the segment, not amplitude
                result /= window.sum() ** 2
        elif mode == "magnitude":
            result = np.abs(result) / window.sum()
        elif mode == "complex":
            result /= window.sum()
        elif mode in ("angle", "phase"):
            # we unwrap the phase later to handle the onesided vs. twosided case
            result = np.angle(result)
            if mode == "phase":
                # we unwrap the phase here to handle the onesided vs. twosided case
                result = np.unwrap(result, axis=0)
        results.append(np.real(result))

    if results:
        lastresult = np.hstack(results)
        t_x = lastresult.shape[0] * lastresult.shape[1]
        t = np.arange(NFFT / 2, t_x - NFFT / 2 + 1, NFFT - noverlap) / Fs
        freqs = fftfreq(pad_to, 1 / Fs)[:numFreqs]
        # center the frequency range at zero
        freqs = np.roll(freqs, -freqcenter, axis=0)
        lastresult = np.roll(lastresult, -freqcenter, axis=0)
        return lastresult, freqs, t
    return (None, None, None)


def specgram(
    x,
    NFFT=None,
    Fs=None,
    Fc=None,
    detrend=None,
    window=None,
    noverlap=None,
    xextent=None,
    pad_to=None,
    scale_by_freq=None,
    mode=None,
    scale=None,
    skip_fft=False,
):
    if NFFT is None:
        NFFT = 256  # same default as in mlab.specgram()
    if Fc is None:
        Fc = 0  # same default as in mlab._spectral_helper()
    if noverlap is None:
        noverlap = 128  # same default as in mlab.specgram()
    if Fs is None:
        Fs = 2  # same default as in mlab._spectral_helper()

    if mode == "complex":
        raise ValueError("Cannot plot a complex specgram")

    if scale is None or scale == "default":
        if mode in ["angle", "phase"]:
            scale = "linear"
        else:
            scale = "dB"
    elif mode in ["angle", "phase"] and scale == "dB":
        raise ValueError("Cannot use dB scale with angle or phase mode")

    spec, freqs, t = spectral_helper(
        x=x,
        NFFT=NFFT,
        Fs=Fs,
        detrend_func=detrend,
        window=window,
        noverlap=noverlap,
        pad_to=pad_to,
        scale_by_freq=scale_by_freq,
        mode=mode,
        skip_fft=skip_fft,
    )
    if scale == "linear":
        Z = spec
    elif scale == "dB":
        if mode is None or mode == "default" or mode == "psd":
            Z = 10.0 * np.log10(spec)
        else:
            Z = 20.0 * np.log10(spec)
    else:
        raise ValueError(f"Unknown scale {scale!r}")

    Z = np.flipud(Z)

    if xextent is None:
        # padding is needed for first and last segment:
        pad_xextent = (NFFT - noverlap) / Fs / 2
        xextent = np.min(t) - pad_xextent, np.max(t) + pad_xextent
    xmin, xmax = xextent
    freqs += Fc
    freqs /= 1e6
    extent = xmin, xmax, freqs[0], freqs[-1]

    return Z, extent


def plot_spectrogram(
    x,
    spectrogram_filename,
    nfft,
    fs,
    fc,
    cmap,
    ytics,
    bare,
    noverlap,
    skip_fft,
    width,
    height,
    dpi,
):
    fig = plt.figure()
    fig.set_size_inches(width, height)
    axes = fig.add_subplot(111)
    axes.set_xlabel("time (s)")
    axes.set_ylabel("freq (MHz)")
    Z, extent = specgram(
        x, NFFT=nfft, Fs=fs, Fc=fc, noverlap=noverlap, skip_fft=skip_fft
    )
    im = axes.imshow(Z, cmap=cmap, extent=extent, origin="upper")
    axes.axis("auto")
    axes.minorticks_on()
    plt.locator_params(axis="y", nbins=ytics)

    if bare:
        axes.set_axis_off()
        plt.subplots_adjust(top=1, bottom=0, right=1, left=0, hspace=0, wspace=0)
        plt.margins(0, 0)
        axes.xaxis.set_major_locator(plt.NullLocator())
        axes.yaxis.set_major_locator(plt.NullLocator())

    plt.sci(im)
    plt.savefig(spectrogram_filename, dpi=dpi)
    print(f"wrote {spectrogram_filename}")
    # must call this in specific order to avoid pyplot leak
    # axes.images.remove(im)
    im.remove()
    fig.clear()
    plt.close()
    plt.cla()
    plt.clf()


def process_recording(args, recording):
    meta = parse_filename(recording)
    center_frequency = meta["center_frequency"]
    sample_rate = int(meta["sample_rate"])
    sample_len = int(meta["sample_len"])
    sample_dtype = meta["sample_dtype"]
    spectrogram_filename = replace_ext(recording, args.iext, all_ext=True)
    if args.skip_exist and os.path.exists(spectrogram_filename):
        print(f"skipping {recording}")
        return
    if is_fft(recording):
        if not args.skip_fft:
            print(f"skipping precomputed FFT {recording}")
            return
    else:
        if args.skip_fft:
            print(f"{recording} not a precomputed FFT, skipping")
            return
    print(f"processing {recording}")
    samples = read_recording(
        recording,
        sample_rate,
        sample_dtype,
        sample_len,
        skip_sample_secs=args.skip_sample_secs,
        max_sample_secs=args.max_sample_secs,
    )
    plot_spectrogram(
        samples,
        spectrogram_filename,
        args.nfft,
        sample_rate,
        center_frequency,
        args.cmap,
        args.ytics,
        args.bare,
        args.noverlap,
        args.skip_fft,
        args.width,
        args.height,
        args.dpi,
    )


def process_all_recordings(args):
    if os.path.isdir(args.recording):
        recordings = get_nondot_files(args.recording)
    else:
        recordings = [args.recording]
    if args.workers == 1:
        for recording in sorted(recordings):
            process_recording(args, recording)
    else:
        with concurrent.futures.ProcessPoolExecutor(
            max_workers=args.workers
        ) as executor:
            for recording in sorted(recordings):
                executor.submit(process_recording, args, recording)
            executor.shutdown(wait=True)


def argument_parser():
    parser = argparse.ArgumentParser(description="draw spectrogram from recording")
    parser.add_argument(
        "recording", default="", type=str, help="filename of recording, or directory"
    )
    parser.add_argument(
        "--nfft", default=int(2048), type=int, help="number of FFT points"
    )
    parser.add_argument("--ytics", default=20, type=int, help="number of y tics")
    # https://ai.googleblog.com/2019/08/turbo-improved-rainbow-colormap-for.html (turbo scheme suggested for ML).
    # yellow signal > blue signal.
    parser.add_argument(
        "--cmap",
        default="turbo_r",
        type=str,
        help="pyplot colormap (see https://matplotlib.org/stable/tutorials/colors/colormaps.html)",
    )
    parser.add_argument("--bare", dest="bare", action="store_true")
    parser.add_argument("--no-bare", dest="bare", action="store_false")
    parser.add_argument(
        "--iext",
        dest="iext",
        default="png",
        type=str,
        help="extension (image type) to use for spectrogram",
    )
    parser.add_argument(
        "--noverlap",
        dest="noverlap",
        default=0,
        type=int,
        help="number of overlapping FFT windows",
    )
    parser.add_argument(
        "--workers",
        dest="workers",
        default=1,
        type=int,
        help="number of parallel workers",
    )
    parser.add_argument(
        "--skip-exist",
        dest="skip_exist",
        action="store_true",
        help="skip existing images",
    )
    parser.add_argument(
        "--no-skip-exist",
        dest="skip_exist",
        action="store_false",
        help="overwrite existing images",
    )
    parser.add_argument(
        "--skip-fft", dest="skip_fft", action="store_true", help="skip FFT"
    )
    parser.add_argument(
        "--no-skip-fft", dest="skip_fft", action="store_false", help="calculate FFT"
    )
    parser.add_argument(
        "--loop", dest="loop", default=0, type=int, help="if > 0, run in a loop"
    )
    parser.add_argument("--width", default=11, type=int, help="plot width")
    parser.add_argument("--height", default=8, type=int, help="plot height")
    parser.add_argument("--dpi", default=100, type=int, help="plot DPI")
    parser.add_argument(
        "--skip_sample_secs",
        default=0,
        type=float,
        help="skip n seconds worth of samples (default skip none)",
    )
    parser.add_argument(
        "--max_sample_secs",
        default=0,
        type=float,
        help="max n seconds worth of samples (default all samples)",
    )
    parser.set_defaults(bare=False, skip_exist=False, skip_fft=False)
    return parser


def main():
    parser = argument_parser()
    args = parser.parse_args()
    while True:
        process_all_recordings(args)
        if not args.loop:
            break
        time.sleep(args.loop)



import argparse
import os
import sys
import time
import tarfile
import sysrsync


def check_tld(top_dir, args):
    if not os.path.exists(top_dir):
        print(f"Top-level directory '{top_dir}' does not exist.")
        return []

    if not os.path.isdir(top_dir):
        print(f"'{top_dir}' is not a directory.")
        return []

    valid_folders = []
    current_time = time.time()

    print(
        f"Folders inside '{top_dir}' that are more than {args.threshold_seconds} seconds old:"
    )

    for sdir in os.listdir(top_dir):
        dir_path = os.path.join(top_dir, sdir)
        if os.path.isdir(dir_path):
            folder_mtime = os.path.getmtime(dir_path)
            if current_time - folder_mtime > args.threshold_seconds:
                print(f"Folder: {dir}")
                valid_folders.append(dir_path)

    return valid_folders


def tar_directories(dir_paths, args):
    # Get a list of subdirectories within the top-level directory
    # subdirectories = [subdir for subdir in os.listdir(top_dir)
    #                  if os.path.isdir(os.path.join(top_dir, subdir))]

    tar_filenames = []

    for dir_path in dir_paths:
        # Create a tar file object
        if args.compress:
            tar_filename = f"{dir_path}.tar.gz"
            file_mode = "w:gz"
        else:
            tar_filename = f"{dir_path}.tar"
            file_mode = "w"

        with tarfile.open(tar_filename, file_mode) as tar_file:
            # Add all files in the directory to the tar file
            for root, _dirs, files in os.walk(dir_path):
                for file in files:
                    file_path = os.path.join(root, file)
                    tar_file.add(file_path)

        # Close the tar file
        tar_file.close()

        tar_filenames.append(tar_filename)

        # Delete if --delete
        if args.delete:
            for root, _dirs, files in os.walk(dir_path):
                for file in files:
                    file_path = os.path.join(root, file)
                    os.remove(file_path)
            os.rmdir(dir_path)

    return tar_filenames


def export_to_path(filename, export_path, args):
    base_filename = os.path.basename(filename)
    exported_filepath = os.path.join(export_path, base_filename)

    if args.export_ssh_host and args.export_ssh_key:
        sysrsync.run(
            source=filename,
            destination=export_path,
            destination_ssh=args.export_ssh_host,
            private_key=args.export_ssh_key,
            options=["-r", "-P", "-u"],
            sync_source_contents=False,
        )
    elif args.export_ssh_host and not args.export_ssh_key:
        sysrsync.run(
            source=filename,
            destination=export_path,
            destination_ssh=args.export_ssh_host,
            options=["-r", "-P", "-u"],
            sync_source_contents=False,
        )
    else:
        print(sysrsync.get_rsync_command(source=filename, destination=export_path))
        sysrsync.run(
            source=filename,
            destination=export_path,
            options=["-r", "-P", "-u"],
            sync_source_contents=False,
        )

    print(f"Exported {filename} to {exported_filepath}")

    return exported_filepath


def argument_parser():
    parser = argparse.ArgumentParser(
        description="tar and compress recording directories"
    )
    parser.add_argument(
        "dir", default="", type=str, help="Top level directory containing recordings"
    )
    parser.add_argument(
        "--compress",
        dest="compress",
        action="store_true",
        default=False,
        help="compress (gzip) directories",
    )
    parser.add_argument(
        "--delete",
        dest="delete",
        action="store_true",
        default=False,
        help="delete after compressing",
    )
    parser.add_argument(
        "--threshold_seconds",
        dest="threshold_seconds",
        type=int,
        default=300,
        help="modtime threshold for folders to be considered, must be older than threshold_seconds",
    )
    parser.add_argument(
        "--export_path",
        dest="export_path",
        type=str,
        default=None,
        help="path to export to after processing",
    )
    parser.add_argument(
        "--export_ssh_host",
        dest="export_ssh_host",
        type=str,
        default=None,
        help="host for external export using ssh",
    )
    parser.add_argument(
        "--export_ssh_key",
        dest="export_ssh_key",
        type=str,
        default=None,
        help="if using external_ssh_host a ssh key can be specified",
    )

    return parser


def check_args(args):
    if args.export_ssh_host and not args.export_ssh_key:
        return "If using export_ssh_key please include export_ssh_host"
    return ""


def main():
    args = argument_parser().parse_args()
    results = check_args(args)
    if results:
        print(results)
        sys.exit(1)

    # Check for valid subdirs
    valid_folders = check_tld(args.dir, args)

    # Process subdirs
    if valid_folders:
        tarred_filenames = tar_directories(valid_folders, args)
        print("Tarred files:")
        for filename in tarred_filenames:
            print(filename)
        print("...finished processing files")
        if args.export_path is not None:
            print(f"Exporting tar files to {args.export_path}")
            export_to_path(args.dir, args.export_path, args)
            print(f"...done exporting to {args.export_path}")
    else:
        print("No valid folders found.")


if __name__ == "__main__":
    main()



#!/usr/bin/python3
import random
from collections import Counter
from collections import defaultdict

import numpy as np
from gamutrflib.peak_finder import get_peak_finder
from gamutrf.utils import SCAN_FRES, SCAN_FROLL


ROLLOVERHZ = 100e6
CSV = ".csv"
ROLLING_FACTOR = int(SCAN_FROLL / SCAN_FRES)


def choose_recorders(signals, recorder_freq_exclusions, max_recorder_signals):
    suitable_recorders = defaultdict(set)
    for signal in sorted(signals):
        for recorder, excluded in sorted(recorder_freq_exclusions.items()):
            if not freq_excluded(signal, excluded):
                suitable_recorders[signal].add(recorder)
    recorder_assignments = []
    busy_count = defaultdict(int)
    for signal, recorders in sorted(suitable_recorders.items(), key=lambda x: x[1]):
        if not recorders:
            continue
        busy_recorders = set(
            recorder
            for recorder, count in busy_count.items()
            if count >= max_recorder_signals
        )
        free_recorders = recorders - busy_recorders
        if not free_recorders:
            continue
        recorder = random.choice(list(free_recorders))  # nosec
        busy_count[recorder] += 1
        recorder_assignments.append((signal, recorder))
    return recorder_assignments


def parse_freq_excluded(freq_exclusions_raw):
    freq_exclusions = []
    for pair in freq_exclusions_raw:
        freq_min, freq_max = pair.split("-")
        if len(freq_min):
            freq_min = int(freq_min)
        else:
            freq_min = None
        if len(freq_max):
            freq_max = int(freq_max)
        else:
            freq_max = None
        freq_exclusions.append((freq_min, freq_max))
    return tuple(freq_exclusions)


def freq_excluded(freq, freq_exclusions):
    for freq_min, freq_max in freq_exclusions:
        if freq_min is not None and freq_max is not None:
            if freq >= freq_min and freq <= freq_max:
                return True
            continue
        if freq_min is None:
            if freq <= freq_max:
                return True
            continue
        if freq >= freq_min:
            return True
    return False


def calc_db(df, rolling_factor=ROLLING_FACTOR):
    df.replace([np.inf, -np.inf], np.nan, inplace=True)
    meandb = df["db"].mean()
    if rolling_factor:
        df["db"] = df["db"].rolling(rolling_factor).mean().fillna(meandb)
    return df


def find_sig_windows(df, detection_type):
    if detection_type:
        peak_finder = get_peak_finder(detection_type)
        if peak_finder:
            db_data = df.db.to_numpy()
            peaks, _ = peak_finder.find_peaks(db_data)
            return [(df.iloc[peak].freq, df.iloc[peak].db) for peak in peaks]
    return []


def get_center(signal_mhz, freq_start_mhz, bin_mhz, record_bw):
    return int(
        int((signal_mhz - freq_start_mhz) / record_bw) * bin_mhz + freq_start_mhz
    )


def choose_record_signal(signals, max_signals):
    recorder_buckets = Counter()

    # Convert signals into buckets of record_bw size, count how many of each size
    for bucket in signals:
        recorder_buckets[bucket] += 1

    # Now count number of buckets of each count.
    buckets_by_count = defaultdict(set)
    for bucket, count in recorder_buckets.items():
        buckets_by_count[count].add(bucket)

    recorder_freqs = []
    # From least occuring bucket to most occurring, choose a random bucket for each recorder.
    for _count, buckets in sorted(buckets_by_count.items()):
        while buckets and len(recorder_freqs) < max_signals:
            bucket = random.choice(list(buckets))  # nosec
            buckets = buckets.remove(bucket)
            recorder_freqs.append(bucket)
        if len(recorder_freqs) == max_signals:
            break
    return recorder_freqs



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import logging
import sys
from pathlib import Path

try:
    from gnuradio import filter as grfilter  # pytype: disable=import-error
    from gnuradio import blocks  # pytype: disable=import-error
    from gnuradio import fft  # pytype: disable=import-error
    from gnuradio import gr  # pytype: disable=import-error
    from gnuradio import zeromq  # pytype: disable=import-error
    from gnuradio.fft import window  # pytype: disable=import-error
except ModuleNotFoundError as err:  # pragma: no cover
    print(
        "Run from outside a supported environment, please run via Docker (https://github.com/IQTLabs/gamutRF#readme): %s"
        % err
    )
    sys.exit(1)

from gamutrf.grsource import get_source
from gamutrf.grinference2mqtt import inference2mqtt
from gamutrf.utils import endianstr


class grscan(gr.top_block):
    def __init__(
        self,
        bucket_range=1.0,
        colormap=16,
        compass=False,
        db_clamp_ceil=50,
        db_clamp_floor=-200,
        dc_block_len=0,
        dc_block_long=False,
        description="",
        external_gps_server="",
        external_gps_server_port=8888,
        fft_batch_size=256,
        freq_end=1e9,
        freq_start=100e6,
        gps_server="",
        igain=0,
        inference_min_confidence=0.5,
        inference_min_db=-200,
        inference_model_server="",
        inference_model_name="",
        inference_output_dir="",
        iqtlabs=None,
        logaddr="0.0.0.0",  # nosec
        logport=8001,
        low_power_hold_down=False,
        mqtt_server="",
        n_image=0,
        n_inference=0,
        nfft=1024,
        pretune=False,
        rotate_secs=0,
        samp_rate=4.096e6,
        sample_dir="",
        scaling="spectrum",
        sdr="ettus",
        sdrargs=None,
        sigmf=True,
        skip_tune_step=0,
        sweep_sec=30,
        tune_step_fft=0,
        tune_dwell_ms=0,
        tuneoverlap=0.5,
        tuning_ranges="",
        use_external_gps=False,
        use_external_heading=False,
        vkfft=False,
        wavelearner=None,
        write_samples=0,
        tag_now=False,
    ):
        gr.top_block.__init__(self, "scan", catch_exceptions=True)

        tune_step_hz = int(samp_rate * tuneoverlap)
        stare = False

        if freq_end == 0:
            stare = True
            freq_end = freq_start + (tune_step_hz - 1)
            logging.info(
                f"using stare mode, scan from {freq_start/1e6}MHz to {freq_end/1e6}MHz"
            )

        ##################################################
        # Parameters
        ##################################################
        self.freq_end = freq_end
        self.freq_start = freq_start
        self.sweep_sec = sweep_sec
        self.nfft = nfft
        self.wavelearner = wavelearner
        self.iqtlabs = iqtlabs
        self.samp_rate = samp_rate
        self.retune_pre_fft = None
        self.tag_now = tag_now

        ##################################################
        # Blocks
        ##################################################

        logging.info(f"will scan from {freq_start} to {freq_end}")
        self.sources, cmd_port, self.workaround_start_hook = get_source(
            sdr,
            samp_rate,
            igain,
            agc=False,
            center_freq=freq_start,
            sdrargs=sdrargs,
        )

        self.samples_blocks = []
        if write_samples:
            Path(sample_dir).mkdir(parents=True, exist_ok=True)
            self.samples_blocks.extend(
                [
                    blocks.complex_to_interleaved_short(False, 32767),
                    blocks.stream_to_vector(gr.sizeof_short, nfft * 2),
                    self.iqtlabs.write_freq_samples(
                        "rx_freq",
                        gr.sizeof_short,
                        "_".join(("ci16", endianstr())),
                        nfft * 2,
                        sample_dir,
                        "samples",
                        write_samples,
                        skip_tune_step,
                        int(samp_rate),
                        rotate_secs,
                        igain,
                        sigmf,
                    ),
                ]
            )
        freq_range = freq_end - freq_start
        fft_rate = int(samp_rate / nfft)

        if not tune_step_fft:
            if tune_dwell_ms:
                tune_step_fft = int(fft_rate * (tune_dwell_ms / 1e3))
            else:
                target_retune_hz = freq_range / self.sweep_sec / tune_step_hz
                tune_step_fft = int(fft_rate / target_retune_hz)
                logging.info(
                    f"retuning across {freq_range/1e6}MHz in {self.sweep_sec}s, requires retuning at {target_retune_hz}Hz in {tune_step_hz/1e6}MHz steps ({tune_step_fft} FFTs)"
                )
        if not tune_step_fft:
            logging.info("tune_step_fft cannot be 0 - defaulting to nfft")
            tune_step_fft = nfft
        tune_dwell_ms = tune_step_fft / fft_rate * 1e3
        logging.info(
            f"requested retuning across {freq_range/1e6}MHz every {tune_step_fft} FFTs, dwell time {tune_dwell_ms}ms"
        )
        if stare and tune_dwell_ms > 1e3:
            logging.warn(">1s dwell time in stare mode, updates will be slow!")

        self.fft_blocks = (
            self.get_dc_blocks(dc_block_len, dc_block_long)
            + self.get_fft_blocks(
                vkfft,
                fft_batch_size,
                nfft,
                freq_start,
                freq_end,
                tune_step_hz,
                tune_step_fft,
                skip_tune_step,
                tuning_ranges,
                pretune,
            )
            + self.get_db_blocks(nfft, samp_rate, scaling)
        )
        fft_dir = sample_dir
        if not write_samples:
            fft_dir = ""
        retune_fft = self.iqtlabs.retune_fft(
            "rx_freq",
            nfft,
            nfft,
            int(samp_rate),
            int(freq_start),
            int(freq_end),
            tune_step_hz,
            tune_step_fft,
            skip_tune_step,
            db_clamp_floor,
            db_clamp_ceil,
            fft_dir,
            write_samples,
            bucket_range,
            tuning_ranges,
            description,
            rotate_secs,
            False,
            self.tag_now,
            not pretune and low_power_hold_down,
        )
        self.fft_blocks.append(retune_fft)
        zmq_addr = f"tcp://{logaddr}:{logport}"
        logging.info("serving FFT on %s", zmq_addr)
        self.fft_blocks.append((zeromq.pub_sink(1, 1, zmq_addr, 100, False, 65536, "")))

        self.inference_blocks = []
        if inference_output_dir:
            x = 640
            y = 640
            Path(inference_output_dir).mkdir(parents=True, exist_ok=True)
            self.inference_blocks.extend(
                [
                    self.iqtlabs.image_inference(
                        tag="rx_freq",
                        vlen=nfft,
                        x=x,
                        y=y,
                        image_dir=inference_output_dir,
                        convert_alpha=255,
                        norm_alpha=0,
                        norm_beta=1,
                        norm_type=32,  # cv::NORM_MINMAX = 32
                        colormap=colormap,  # cv::COLORMAP_VIRIDIS = 16, cv::COLORMAP_TURBO = 20,
                        interpolation=1,  # cv::INTER_LINEAR = 1,
                        flip=0,  # 0 means flipping around the x-axis
                        min_peak_points=inference_min_db,
                        model_server=inference_model_server,
                        model_name=inference_model_name,
                        confidence=inference_min_confidence,
                        max_rows=tune_step_fft,
                        rotate_secs=rotate_secs,
                        n_image=n_image,
                        n_inference=n_inference,
                    )
                ]
            )
            if mqtt_server:
                self.inference_blocks.extend(
                    [
                        inference2mqtt(
                            "inference2mqtt",
                            mqtt_server,
                            compass,
                            gps_server,
                            use_external_gps,
                            use_external_heading,
                            external_gps_server,
                            external_gps_server_port,
                            inference_output_dir,
                        )
                    ]
                )
            else:
                self.inference_blocks.extend([blocks.null_sink(1)])
        if not self.inference_blocks:
            self.inference_blocks = [blocks.null_sink(gr.sizeof_float * nfft)]

        if pretune:
            self.msg_connect((self.retune_pre_fft, "tune"), (self.sources[0], cmd_port))
            self.msg_connect((self.retune_pre_fft, "tune"), (retune_fft, "cmd"))
        else:
            self.msg_connect((retune_fft, "tune"), (self.sources[0], cmd_port))
        self.connect_blocks(self.sources[0], self.sources[1:])
        self.connect((retune_fft, 1), (self.inference_blocks[0], 0))

        self.connect_blocks(self.inference_blocks[0], self.inference_blocks[1:])
        for pipeline_blocks in (
            self.fft_blocks,
            self.samples_blocks,
        ):
            self.connect_blocks(self.sources[-1], pipeline_blocks)

    def connect_blocks(self, source, other_blocks, last_block_port=0):
        last_block = source
        for block in other_blocks:
            self.connect((last_block, last_block_port), (block, 0))
            last_block = block

    def get_db_blocks(self, nfft, samp_rate, scaling):
        if scaling == "density":
            scale = 1.0 / (samp_rate * sum(self.get_window(nfft)) ** 2)
        elif scaling == "spectrum":
            scale = 1.0 / (sum(self.get_window(nfft)) ** 2)
        else:
            raise ValueError("scaling must be 'spectrum' or 'density'")
        return [
            blocks.complex_to_mag_squared(nfft),
            blocks.multiply_const_ff(scale, nfft),
            blocks.nlog10_ff(10, nfft, 0),
        ]

    def get_window(self, nfft):
        return window.hann(nfft)

    def get_pretune_block(
        self,
        fft_batch_size,
        nfft,
        freq_start,
        freq_end,
        tune_step_hz,
        tune_step_fft,
        skip_tune_step,
        tuning_ranges,
        pretune,
    ):
        # if pretuning, the pretune block will also do the batching.
        if pretune:
            return self.iqtlabs.retune_pre_fft(
                nfft,
                fft_batch_size,
                "rx_freq",
                int(freq_start),
                int(freq_end),
                tune_step_hz,
                tune_step_fft,
                skip_tune_step,
                tuning_ranges,
                self.tag_now,
            )
        # otherwise, the pretuning block will just do batching.
        return blocks.stream_to_vector(gr.sizeof_gr_complex, fft_batch_size * nfft)

    def get_offload_fft_blocks(
        self,
        vkfft,
        fft_batch_size,
        nfft,
    ):
        fft_block = None
        fft_roll = False
        if self.wavelearner:
            fft_block = self.wavelearner.fft(int(fft_batch_size * nfft), (nfft), True)
            fft_roll = True
        elif vkfft:
            fft_block = self.iqtlabs.vkfft(int(fft_batch_size * nfft), nfft, True)
        else:
            fft_batch_size = 1
            fft_blocks = [
                fft.fft_vcc(nfft, True, self.get_window(nfft), True, 1),
            ]
            return fft_batch_size, fft_blocks

        fft_blocks = [
            blocks.multiply_const_vff(
                [val for val in self.get_window(nfft) for _ in range(2)]
                * fft_batch_size
            ),
            fft_block,
            blocks.vector_to_stream(gr.sizeof_gr_complex * nfft, fft_batch_size),
        ]
        if fft_roll:
            fft_blocks.append(self.iqtlabs.vector_roll(nfft))
        return fft_batch_size, fft_blocks

    def get_dc_blocks(self, dc_block_len, dc_block_long):
        if dc_block_len:
            return [grfilter.dc_blocker_cc(dc_block_len, dc_block_long)]
        return []

    def get_fft_blocks(
        self,
        vkfft,
        fft_batch_size,
        nfft,
        freq_start,
        freq_end,
        tune_step_hz,
        tune_step_fft,
        skip_tune_step,
        tuning_ranges,
        pretune,
    ):
        fft_batch_size, fft_blocks = self.get_offload_fft_blocks(
            vkfft, fft_batch_size, nfft
        )
        self.retune_pre_fft = self.get_pretune_block(
            fft_batch_size,
            nfft,
            freq_start,
            freq_end,
            tune_step_hz,
            tune_step_fft,
            skip_tune_step,
            tuning_ranges,
            pretune,
        )
        return [self.retune_pre_fft] + fft_blocks

    def start(self):
        super().start()
        self.workaround_start_hook(self)



#!/usr/bin/python3
import logging
import os
import re
import sys
from pathlib import Path

import requests

MTU = 8962
SCAN_FRES = 1e4
SCAN_FROLL = 1e6

# Use max recv_frame_size for USB - because we don't mind latency,
# we are optimizing for lower CPU.
# https://files.ettus.com/manual/page_transport.html
# https://github.com/EttusResearch/uhd/blob/master/host/lib/usrp/b200/b200_impl.hpp
# Should result in no overflows:
# UHD_IMAGES_DIR=/usr/share/uhd/images ./examples/rx_samples_to_file --args num_recv_frames=1000,recv_frame_size=16360 --file test.gz --nsamps 200000000 --rate 20000000 --freq 101e6 --spb 20000000
ETTUS_ARGS = "num_recv_frames=1000,recv_frame_size=16360,type=b200"
ETTUS_ANT = "TX/RX"
SAMPLE_FILENAME_RE = re.compile(r"^.+\D(\d+)_(\d+)Hz.*\D(\d+)sps\.c*([fisu]\d+|raw).*$")

SAMPLE_DTYPES = {
    "i8": ("<i1", "signed-integer"),
    "i16": ("<i2", "signed-integer"),
    "i32": ("<i4", "signed-integer"),
    "s8": ("<i1", "signed-integer"),
    "s16": ("<i2", "signed-integer"),
    "s32": ("<i4", "signed-integer"),
    "u8": ("<u1", "unsigned-integer"),
    "u16": ("<u2", "unsigned-integer"),
    "u32": ("<u4", "unsigned-integer"),
    "f32": ("<u4", "unsigned-integer"),
    "raw": ("<f4", "float"),
}
WIDTH = 11
HEIGHT = 8
DPI = 75
MPL_BACKEND = "cairo"
DS_PIXELS = 256
SAMP_RATE = 8.192e6
MIN_FREQ = 2.3e9
MAX_FREQ = 2.6e9


def endianstr():
    if sys.byteorder == "little":
        return "le"
    return "be"


def rotate_file_n(initial_name, n, require_initial=True):
    if require_initial and not os.path.exists(initial_name):
        return
    base_name = initial_name
    ext = ""
    dot = initial_name.rfind(".")
    if dot != -1:
        ext = base_name[dot:]
        base_name = base_name[:dot]
    for i in range(n, 1, -1):
        from_name = f"{base_name}.{i-1}{ext}"
        to_name = f"{base_name}.{i}{ext}"
        if os.path.exists(from_name):
            os.rename(from_name, to_name)
    if require_initial:
        os.rename(initial_name, f"{base_name}.1{ext}")


def replace_ext(filename, ext, all_ext=False):
    basename = os.path.basename(filename)
    if all_ext:
        dot = basename.index(".")
    else:
        dot = basename.rindex(".")
    new_basename = basename[: (dot + 1)] + ext
    return filename.replace(basename, new_basename)


def is_fft(filename):
    return os.path.basename(filename).startswith("fft_")


def get_nondot_files(filedir, glob="*.s*.*"):
    return [
        str(path)
        for path in Path(filedir).rglob(glob)
        if not os.path.basename(path).startswith(".")
    ]


def http_get(url, timeout=10):
    try:
        req = requests.get(url, timeout=timeout)
        logging.debug(str(req))
        return req
    except (requests.exceptions.ConnectionError, requests.exceptions.Timeout) as err:
        logging.error(f"{url}: {str(err)}")
        return None



import json
import logging
import os
import socket
import time

import gpsd
import paho.mqtt.client as mqtt
from gamutrf.utils import http_get


class MQTTReporter:
    def __init__(
        self,
        name,
        mqtt_server=None,
        gps_server=None,
        compass=False,
        use_external_gps=False,
        use_external_heading=False,
        external_gps_server=None,
        external_gps_server_port=None,
    ):
        self.name = name
        self.mqtt_server = mqtt_server
        self.compass = compass
        self.gps_server = gps_server
        self.mqttc = None
        self.heading = "no heading"
        self.use_external_gps = use_external_gps
        self.use_external_heading = use_external_heading
        self.external_gps_server = external_gps_server
        self.external_gps_server_port = external_gps_server_port
        self.external_gps_msg = None
        self.gps_configured = True
        if not self.gps_server and not self.external_gps_server:
            logging.error("mqtt enabled, no gps_server or external_gps_server found")
            self.gps_configured = False
        if (
            self.external_gps_server
            and not self.use_external_gps
            and not self.gps_server
        ):
            logging.error(
                "mqtt enabled and only external_gps_server found, but no use_external_gps flag"
            )
            self.gps_configured = False

    @staticmethod
    def log(path, prefix, start_time, record_args):
        try:
            with open(
                os.path.join(path, f"mqtt-{prefix}-{start_time}.log"),
                "a+",
                encoding="utf-8",
            ) as f:
                f.write(f"{json.dumps(record_args)}\n")
        except FileNotFoundError as err:
            logging.error(f"could not write to mqtt rssi log: {err}")

    def connect(self):
        logging.info(f"connecting to {self.mqtt_server}")
        self.mqttc = mqtt.Client()
        self.mqttc.connect(self.mqtt_server)
        self.mqttc.loop_start()

    def get_heading(self):
        if self.use_external_heading:
            heading_result = http_get(
                f"http://{self.external_gps_server}:{self.external_gps_server_port}/heading"
            )
            if heading_result is None:
                logging.error("could not update external heading")
            else:
                self.heading = float(json.loads(heading_result.text)["heading"])
        else:
            heading_result = http_get(f"http://{self.gps_server}:8000/v1/heading")
            if heading_result is None:
                logging.error("could not update heading")
            else:
                self.heading = str(float(heading_result.text))

    def add_gps(self, publish_args):
        if not self.gps_configured:
            return publish_args
        publish_args.update(
            {
                "position": [0, 0],
                "altitude": None,
                "gps_time": None,
                "map_url": None,
                "heading": self.heading,
                "gps": "no fix",
            }
        )

        # Use external external GPS
        if self.use_external_gps:
            external_gps_msg = http_get(
                f"http://{self.external_gps_server}:{self.external_gps_server_port}/gps-data"
            )
            if external_gps_msg is None:
                logging.error("could not update with external GPS")
            else:
                self.external_gps_msg = json.loads(external_gps_msg.text)
                publish_args.update(
                    {
                        "position": (
                            self.external_gps_msg["latitude"],
                            self.external_gps_msg["longitude"],
                        ),
                        "altitude": self.external_gps_msg["altitude"],
                        "gps_time": self.external_gps_msg["time_usec"],
                        "map_url": None,
                        "heading": self.heading,
                        "gps": "fix",
                    }
                )

        # Use internal GPIO GPS
        else:
            try:
                if self.compass:
                    self.get_heading()
                if gpsd.gpsd_stream is None:
                    gpsd.connect(host=self.gps_server, port=2947)
                packet = gpsd.get_current()
                publish_args.update(
                    {
                        "position": packet.position(),
                        "altitude": packet.altitude(),
                        "gps_time": packet.get_time().timestamp(),
                        "map_url": packet.map_url(),
                        "heading": self.heading,
                        "gps": "fix",
                    }
                )
            except (BrokenPipeError, gpsd.NoFixError, AttributeError) as err:
                logging.error("could not update with GPS: %s", err)
        return publish_args

    def publish(self, publish_path, publish_args):
        if not self.mqtt_server:
            return
        try:
            if self.mqttc is None:
                self.connect()
            publish_args = self.add_gps(publish_args)
            publish_args["name"] = self.name
            self.mqttc.publish(publish_path, json.dumps(publish_args))
        except (
            socket.gaierror,
            ConnectionRefusedError,
            mqtt.WebsocketConnectionError,
            ValueError,
        ) as err:
            logging.error(f"failed to publish to MQTT {self.mqtt_server}: {err}")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import sys
import numpy as np

try:
    from gnuradio import gr  # pytype: disable=import-error
except ModuleNotFoundError as err:  # pragma: no cover
    print(
        "Run from outside a supported environment, please run via Docker (https://github.com/IQTLabs/gamutRF#readme): %s"
        % err
    )
    sys.exit(1)


class terminal_sink(gr.sync_block):
    # Prints output layer of classifier to terminal for troubleshooting
    def __init__(self, input_vlen, batch_size):
        self.input_vlen = input_vlen
        self.batch_size = batch_size
        gr.sync_block.__init__(
            self,
            name="terminal_sink",
            in_sig=[(np.float32, self.input_vlen)],
            out_sig=None,
        )
        self.batch_ctr = 0

    def work(self, input_items, output_items):
        in0 = input_items[0]
        _batch = in0.reshape(self.batch_size, -1)
        self.batch_ctr += 1
        return len(input_items[0])



import copy
import threading
from flask import Flask, request


class FlaskHandler:
    def __init__(self, options, check_options, banned_args):
        self.check_options = check_options
        self.options = options
        self.banned_args = banned_args
        self.reconfigures = 0
        self.app = Flask(__name__)
        self.app.add_url_rule("/reconf", "reconf", self.reconf)
        self.request = request
        self.thread = threading.Thread(
            target=self.app.run,
            kwargs={"host": "0.0.0.0", "port": options.apiport},  # nosec
            daemon=True,
        )

    def start(self):
        self.thread.start()

    def reconf(self):
        new_options = copy.deepcopy(self.options)
        for arg, val in self.request.args.items():
            if arg in self.banned_args:
                continue
            if not hasattr(new_options, arg):
                return f"no such option {arg}", 400
            val_type = type(getattr(self.options, arg))
            try:
                setattr(new_options, arg, val_type(val))
            except (TypeError, ValueError) as err:
                return f"cannot set {arg} = {val}: {err}", 400
        results = self.check_options(new_options)
        if results:
            return results, 400
        self.options = new_options
        self.reconfigures += 1
        return "reconf", 200



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import json
import queue
import sys
import threading
import time
import numpy as np

try:
    from gnuradio import gr  # pytype: disable=import-error
    from gamutrf.mqtt_reporter import MQTTReporter
except ModuleNotFoundError as err:  # pragma: no cover
    print(
        "Run from outside a supported environment, please run via Docker (https://github.com/IQTLabs/gamutRF#readme): %s"
        % err
    )
    sys.exit(1)


DELIM = "\n\n"


class inference2mqtt(gr.sync_block):
    def __init__(
        self,
        name,
        mqtt_server,
        compass,
        gps_server,
        use_external_gps,
        use_external_heading,
        external_gps_server,
        external_gps_server_port,
        log_path,
    ):
        self.yaml_buffer = ""
        self.mqtt_reporter = None
        self.q = queue.Queue()
        self.mqtt_reporter_thread = threading.Thread(
            target=self.reporter_thread,
            args=(
                name,
                mqtt_server,
                gps_server,
                compass,
                use_external_gps,
                use_external_heading,
                external_gps_server,
                external_gps_server_port,
                log_path,
            ),
        )
        self.mqtt_reporter_thread.start()

        gr.sync_block.__init__(
            self,
            name="inference2mqtt",
            in_sig=[np.ubyte],
            out_sig=None,
        )

    def reporter_thread(
        self,
        name,
        mqtt_server,
        gps_server,
        compass,
        use_external_gps,
        use_external_heading,
        external_gps_server,
        external_gps_server_port,
        log_path,
    ):
        start_time = time.time()
        mqtt_reporter = MQTTReporter(
            name=name,
            mqtt_server=mqtt_server,
            gps_server=gps_server,
            compass=compass,
            use_external_gps=use_external_gps,
            use_external_heading=use_external_heading,
            external_gps_server=external_gps_server,
            external_gps_server_port=external_gps_server_port,
        )
        while True:
            item = self.q.get()
            mqtt_reporter.publish("gamutrf/inference", item)
            mqtt_reporter.log(log_path, "inference", start_time, item)
            self.q.task_done()

    def work(self, input_items, output_items):
        n = 0
        for input_item in input_items:
            raw_input_item = input_item.tobytes().decode("utf8")
            n += len(raw_input_item)
            self.yaml_buffer += raw_input_item
        while True:
            delim_pos = self.yaml_buffer.find(DELIM)
            if delim_pos == -1:
                break
            raw_item = self.yaml_buffer[:delim_pos]
            item = json.loads(raw_item)
            self.yaml_buffer = self.yaml_buffer[delim_pos + len(DELIM) :]
            self.process_item(item)
        return n

    def process_item(self, item):
        self.q.put(item)



import sys

try:
    from gnuradio import blocks
    from gnuradio import gr
    from gnuradio import network
except ModuleNotFoundError as err:
    print(
        "Run from outside a supported environment, please run via Docker (https://github.com/IQTLabs/gamutRF#readme): %s"
        % err
    )
    sys.exit(1)

from gamutrf.grsource import get_source

FLOAT_SIZE = 4
RSSI_UDP_ADDR = "127.0.0.1"
RSSI_UDP_PORT = 2001
MAX_RSSI = 100


class BirdsEyeRSSI(gr.top_block):
    def __init__(self, args, samp_rate, center_freq, rssi_throttle=10, agc=False):
        gr.top_block.__init__(self, "BirdsEyeRSSI", catch_exceptions=True)

        self.threshold = args.rssi_threshold
        self.mean_window = args.mean_window
        self.rssi_throttle = rssi_throttle
        sources, _cmd_port, _workaround_start_hook = get_source(
            args.sdr, samp_rate, args.gain, agc, center_freq
        )

        rssi_blocks = sources + [
            blocks.complex_to_mag_squared(1),
            blocks.moving_average_ff(self.mean_window, 1 / self.mean_window, 2000, 1),
            blocks.nlog10_ff(10, 1, 0),
            blocks.add_const_ff(-34),
            blocks.keep_one_in_n(gr.sizeof_float, int(self.rssi_throttle)),
            network.udp_sink(
                gr.sizeof_float, 1, RSSI_UDP_ADDR, RSSI_UDP_PORT, 0, 32768, False
            ),
        ]

        last_block = rssi_blocks[0]
        for block in rssi_blocks[1:]:
            self.connect((last_block, 0), (block, 0))
            last_block = block



import time
from gamutrf.utils import ETTUS_ANT
from gamutrf.utils import ETTUS_ARGS


def get_ettus_source(sdrargs, samp_rate, center_freq, agc, gain, uhd):
    if not sdrargs:
        sdrargs = ETTUS_ARGS
    source = uhd.usrp_source(
        ",".join((sdrargs, "")),
        uhd.stream_args(
            cpu_format="fc32",
            args="",
            channels=list(range(0, 1)),
        ),
    )
    source.set_antenna(ETTUS_ANT, 0)
    source.set_samp_rate(samp_rate)
    if center_freq is not None:
        source.set_center_freq(center_freq, 0)
    source.set_gain(gain, 0)
    source.set_rx_agc(agc, 0)

    now = time.time()
    now_sec = int(now)
    now_frac = now - now_sec
    uhd_now = uhd.time_spec(now_sec, now_frac)
    source.set_time_now(uhd_now, uhd.ALL_MBOARDS)
    time_diff = source.get_time_now().get_real_secs() - now
    if time_diff > 5:
        raise ValueError("could not set Ettus SDR time successfully")
    return [source]



"""Main entrypoint for GamutRF"""
from gamutrf.compress_dirs import main as compress_dirs_main
from gamutrf.offline import main as offline_main
from gamutrf.scan import main as scan_main
from gamutrf.sigfinder import main as sigfinder_main
from gamutrf.specgram import main as specgram_main
from gamutrf.worker import main as worker_main


def compress_dirs():
    """Entrypoint for compress_dirs"""
    compress_dirs_main()


def offline():
    """Entrypoint for offline"""
    offline_main()


def scan():
    """Entrypoint for scan"""
    scan_main()


def sigfinder():
    """Entrypoint for sigfinder"""
    sigfinder_main()


def specgram():
    """Entrypoint for specgram"""
    specgram_main()


def worker():
    """Entrypoint for worker"""
    worker_main()



import argparse
import concurrent.futures
import json
import logging
import os
import subprocess
import threading
import time

import bjoern
import falcon
import jinja2
import numpy as np
import schedule

from prometheus_client import Counter
from prometheus_client import Gauge
from prometheus_client import start_http_server

from gamutrf.sigwindows import calc_db
from gamutrf.sigwindows import choose_record_signal
from gamutrf.sigwindows import choose_recorders
from gamutrf.sigwindows import get_center
from gamutrf.sigwindows import parse_freq_excluded
from gamutrf.sigwindows import find_sig_windows
from gamutrf.sigwindows import ROLLING_FACTOR
from gamutrf.utils import http_get, rotate_file_n, SCAN_FRES
from gamutrflib.zmqbucket import ZmqReceiver, parse_scanners


MB = int(1.024e6)
PEAK_TRIGGER = int(os.environ.get("PEAK_TRIGGER", "0"))
PIN_TRIGGER = int(os.environ.get("PIN_TRIGGER", "17"))
if PEAK_TRIGGER == 1:
    import RPi.GPIO as GPIO

    GPIO.setmode(GPIO.BCM)
    GPIO.setup(PIN_TRIGGER, GPIO.OUT)

PEAK_DBS = {}


def falcon_response(resp, text, status):
    resp.status = status
    resp.text = text
    resp.content_type = "text/html"


def ok_response(resp, text="ok!"):
    falcon_response(resp, text=text, status=falcon.HTTP_200)


def error_response(resp, text="error!"):
    falcon_response(resp, text=text, status=falcon.HTTP_500)


def load_template(name):
    path = os.path.join("templates", name)
    with open(os.path.abspath(path), "r", encoding="utf-8") as fp:
        return jinja2.Template(fp.read())


class ActiveRequests:
    def on_get(self, req, resp):
        all_jobs = schedule.get_jobs()
        ok_response(resp, f"{all_jobs}")


class ScannerForm:
    def on_get(self, req, resp):
        template = load_template("scanner_form.html")
        ok_response(resp, template.render(bins=PEAK_DBS))


class Result:
    def on_post(self, req, resp):
        # TODO validate input
        try:
            recorder = f'http://{req.media["worker"]}:8000/'
            signal_hz = int(int(req.media["frequency"]) * 1e6)
            record_bps = int(int(req.media["bandwidth"]) * MB)
            record_samples = int(record_bps * int(req.media["duration"]))
            action = req.media["action"]
            recorder_args = f"{action}/{signal_hz}/{record_samples}/{record_bps}"
            timeout = int(req.media["duration"])
            response = None
            if int(req.media["repeat"]) == -1:
                schedule.every(timeout).seconds.do(
                    run_threaded,
                    record,
                    recorder=recorder,
                    recorder_args=recorder_args,
                    timeout=timeout,
                ).tag(f"{recorder}{recorder_args}-{timeout}")
                ok_response(resp)
            else:
                response = recorder_req(recorder, recorder_args, timeout)
                time.sleep(timeout)
                for _ in range(int(req.media["repeat"])):
                    response = recorder_req(recorder, recorder_args, timeout)
                    time.sleep(timeout)
                if response:
                    ok_response(resp)
                else:
                    ok_response(resp, f"Request {recorder} {recorder_args} failed.")
        except Exception as e:
            error_response(resp, f"{e}")


def record(recorder, recorder_args, timeout):
    recorder_req(recorder, recorder_args, timeout)


def run_threaded(job_func, recorder, recorder_args, timeout):
    job_thread = threading.Thread(
        target=job_func,
        args=(
            recorder,
            recorder_args,
            timeout,
        ),
    )
    job_thread.start()


def init_prom_vars():
    prom_vars = {
        "last_bin_freq_time": Gauge(
            "last_bin_freq_time",
            "epoch time last signal in each bin",
            labelnames=("bin_mhz",),
        ),
        "worker_record_request": Gauge(
            "worker_record_request",
            "record requests made to workers",
            labelnames=("worker",),
        ),
        "freq_power": Gauge(
            "freq_power", "bin frequencies and db over time", labelnames=("bin_freq",)
        ),
        "new_bins": Counter(
            "new_bins", "frequencies of new bins", labelnames=("bin_freq",)
        ),
        "old_bins": Counter(
            "old_bins", "frequencies of old bins", labelnames=("bin_freq",)
        ),
        "bin_freq_count": Counter(
            "bin_freq_count", "count of signals in each bin", labelnames=("bin_mhz",)
        ),
        "frame_counter": Counter("frame_counter", "number of frames processed"),
    }
    return prom_vars


def update_prom_vars(peak_dbs, new_bins, old_bins, prom_vars):
    freq_power = prom_vars["freq_power"]
    new_bins_prom = prom_vars["new_bins"]
    old_bins_prom = prom_vars["old_bins"]
    for freq in peak_dbs:
        freq_power.labels(bin_freq=freq).set(peak_dbs[freq])
    for nbin in new_bins:
        new_bins_prom.labels(bin_freq=nbin).inc()
    for obin in old_bins:
        old_bins_prom.labels(bin_freq=obin).inc()


def process_scan(args, scan_configs, prom_vars, df, lastbins):
    global PEAK_DBS
    df = calc_db(df, args.db_rolling_factor)
    freqdiffs = df.freq - df.freq.shift()
    mindiff = freqdiffs.min()
    maxdiff = freqdiffs.max()
    meandiff = freqdiffs.mean()
    logging.info(
        "new frame with %u samples, frequency sample differences min %f mean %f max %f",
        len(df),
        mindiff,
        meandiff,
        maxdiff,
    )
    if meandiff > mindiff * 2:
        logging.warning(
            "mean frequency diff larger than minimum - increase scanner sample rate"
        )
        logging.warning(df[freqdiffs > mindiff * 2])
    if args.fftlog:
        tmp_fftlog = os.path.join(
            os.path.dirname(args.fftlog), "." + os.path.basename(args.fftlog)
        )
        df.to_csv(tmp_fftlog, sep="\t", header=False, index=False)
        os.rename(tmp_fftlog, args.fftlog)
    monitor_bins = set()
    peak_dbs = {}
    bin_freq_count = prom_vars["bin_freq_count"]
    last_bin_freq_time = prom_vars["last_bin_freq_time"]
    freq_start_mhz = (
        min([scan_config["freq_start"] for scan_config in scan_configs]) / 1e6
    )
    signals = []
    for scan_config in scan_configs:
        scan_df = df[
            (df.freq >= scan_config["freq_start"] / 1e6)
            & (df.freq <= scan_config["freq_end"] / 1e6)
        ]
        logging.info(args.detection_type)
        if args.detection_type:
            signals.extend(
                find_sig_windows(
                    scan_df,
                    args.detection_type,
                )
            )
    min_samp_rate = (
        min([scan_config["sample_rate"] for scan_config in scan_configs]) / 1e6
    )
    df.loc[freqdiffs > min_samp_rate * 2, "db"] = np.nan

    if PEAK_TRIGGER == 1 and signals:
        led_sleep = 0.2
        GPIO.output(PIN_TRIGGER, GPIO.HIGH)
        time.sleep(led_sleep)
        GPIO.output(PIN_TRIGGER, GPIO.LOW)

    ts = df["ts"].max()
    for peak_freq, peak_db in signals:
        center_freq = get_center(
            peak_freq, freq_start_mhz, args.bin_mhz, args.record_bw_msps
        )
        logging.info(
            "detected peak at %f MHz %f dB, assigned bin frequency %f MHz",
            peak_freq,
            peak_db,
            center_freq,
        )
        bin_freq_count.labels(bin_mhz=center_freq).inc()
        last_bin_freq_time.labels(bin_mhz=ts).set(ts)
        monitor_bins.add(center_freq)
        peak_dbs[center_freq] = peak_db
    logging.info(
        "current bins %f to %f MHz: %s",
        df["freq"].min(),
        df["freq"].max(),
        sorted(peak_dbs.items()),
    )
    PEAK_DBS = sorted(peak_dbs.items())
    new_bins = monitor_bins - lastbins
    if new_bins:
        logging.info("new bins: %s", sorted(new_bins))
    old_bins = lastbins - monitor_bins
    if old_bins:
        logging.info("old bins: %s", sorted(old_bins))
    update_prom_vars(peak_dbs, new_bins, old_bins, prom_vars)
    return monitor_bins


def recorder_req(recorder, recorder_args, timeout):
    return http_get(f"{recorder}/v1/{recorder_args}", timeout=timeout)


def get_freq_exclusions(args):
    recorder_freq_exclusions = {}
    for recorder in args.recorder:
        req = recorder_req(recorder, "info", args.record_secs)
        if req is None or req.status_code != 200:
            continue
        excluded = json.loads(req.text).get("freq_excluded", None)
        if excluded is None:
            continue
        recorder_freq_exclusions[recorder] = parse_freq_excluded(excluded)
    return recorder_freq_exclusions


def call_record_signals(args, lastbins_history, prom_vars):
    if lastbins_history:
        signals = []
        for bins in lastbins_history:
            signals.extend(list(bins))
        recorder_freq_exclusions = get_freq_exclusions(args)
        recorder_count = len(recorder_freq_exclusions)
        record_signals = choose_record_signal(
            signals, recorder_count * args.max_recorder_signals
        )
        for signal, recorder in choose_recorders(
            record_signals, recorder_freq_exclusions, args.max_recorder_signals
        ):
            signal_hz = int(signal * 1e6)
            record_bps = int(args.record_bw_msps * MB)
            record_samples = int(record_bps * args.record_secs)
            recorder_args = f"record/{signal_hz}/{record_samples}/{record_bps}"
            resp = recorder_req(recorder, recorder_args, args.record_secs)
            if resp:
                worker_record_request = prom_vars["worker_record_request"]
                worker_record_request.labels(worker=recorder).set(signal_hz)


def zstd_file(uncompressed_file):
    subprocess.check_call(["/usr/bin/zstd", "--force", "--rm", uncompressed_file])


class ScanProcessor:
    def __init__(self, args, prom_vars):
        self.args = args
        self.prom_vars = prom_vars
        self.lastbins_history = []
        self.lastbins = set()
        self.frame_counter = self.prom_vars["frame_counter"]


def process_scans(args, prom_vars, executor, zmqr):
    sp = ScanProcessor(args, prom_vars)

    while True:
        if os.path.exists(args.log):
            logging.info(f"{args.log} exists, will append first")
            mode = "a"
        else:
            logging.info(f"opening {args.log}")
            mode = "w"
        deadline = int(time.time()) + args.rotatesecs
        with open(args.log, mode=mode, encoding="utf-8") as l:
            while True:
                now = int(time.time())
                if not zmqr.healthy():
                    return
                if now > deadline:
                    break
                scan_configs, frame_df = zmqr.read_buff(l, scan_fres=SCAN_FRES)
                if frame_df is None:
                    schedule.run_pending()
                    sleep_time = 1
                    time.sleep(sleep_time)
                    continue
                sp.frame_counter.inc()
                logging.info(
                    "frame with sweep_start %us ago",
                    now - frame_df["sweep_start"].min(),
                )
                new_lastbins = process_scan(
                    args,
                    scan_configs,
                    prom_vars,
                    frame_df,
                    sp.lastbins,
                )
                if new_lastbins is not None:
                    sp.lastbins = new_lastbins
                    if sp.lastbins:
                        sp.lastbins_history = [sp.lastbins] + sp.lastbins_history
                        sp.lastbins_history = sp.lastbins_history[: args.history]
                call_record_signals(args, sp.lastbins_history, prom_vars)
        rotate_file_n(".".join((args.log, "zst")), args.nlog, require_initial=False)
        new_log = ".".join((args.log, "1"))
        os.rename(args.log, new_log)
        executor.submit(zstd_file, new_log)


def argument_parser():
    parser = argparse.ArgumentParser(
        description="watch a scan UDP stream and find signals"
    )
    parser.add_argument(
        "--log", default="scan.log", type=str, help="base path for scan logging"
    )
    parser.add_argument(
        "--fftlog",
        default="",
        type=str,
        help="if defined, path to log last complete FFT frame",
    )
    parser.add_argument(
        "--rotatesecs",
        default=3600,
        type=int,
        help="rotate scan log after this many seconds",
    )
    parser.add_argument(
        "--nlog", default=10, type=int, help="keep only this many scan.logs"
    )
    parser.add_argument(
        "--bin_mhz", default=20, type=int, help="monitoring bin width in MHz"
    )
    parser.add_argument(
        "--history",
        default=5,
        type=int,
        help="number of frames of signal history to keep",
    )
    parser.add_argument(
        "--recorder",
        action="append",
        default=[],
        help="SDR recorder base URLs (e.g. http://host:port/, multiples can be specified)",
    )
    parser.add_argument(
        "--record_bw_msps",
        default=20,
        type=int,
        help="record bandwidth in n * {MB} samples per second",
    )
    parser.add_argument(
        "--record_secs", default=10, type=int, help="record time duration in seconds"
    )
    parser.add_argument(
        "--promport",
        dest="promport",
        type=int,
        default=9000,
        help="Prometheus client port",
    )
    parser.add_argument(
        "--port",
        dest="port",
        type=int,
        default=80,
        help="control webserver port",
    )
    parser.add_argument(
        "--scanners",
        dest="scanners",
        type=str,
        default="127.0.0.1:8001",
        help="Connect to gamutRF scanners at these addresses",
    )
    parser.add_argument(
        "--max_recorder_signals",
        dest="max_recorder_signals",
        type=int,
        default=1,
        help="Max number of recordings per worker to request",
    )
    parser.add_argument(
        "--buff_path",
        dest="buff_path",
        type=str,
        default="/dev/shm",  # nosec
        help="Path for FFT buffer file",
    )
    parser.add_argument(
        "--db_rolling_factor",
        dest="db_rolling_factor",
        type=float,
        default=0,
        help="Divisor for rolling dB average (or 0 to disable)",
    )
    parser.add_argument(
        "--detection_type",
        default="narrowband",
        type=str,
        help="Detection type to plot (wideband, narrowband).",
    )
    return parser


def main():
    parser = argument_parser()
    args = parser.parse_args()
    scanners = parse_scanners(args.scanners)

    logging.basicConfig(level=logging.DEBUG, format="%(asctime)s %(message)s")
    prom_vars = init_prom_vars()
    app = falcon.App()
    scanner_form = ScannerForm()
    result = Result()
    active_requests = ActiveRequests()
    app.add_route("/", scanner_form)
    app.add_route("/result", result)
    app.add_route("/requests", active_requests)
    start_http_server(args.promport)

    with concurrent.futures.ProcessPoolExecutor(2) as executor:
        zmqr = ZmqReceiver(
            scanners=scanners,
            buff_path=args.buff_path,
        )
        x = threading.Thread(
            target=process_scans,
            args=(
                args,
                prom_vars,
                executor,
                zmqr,
            ),
        )
        x.start()
        try:
            bjoern.run(app, "0.0.0.0", args.port)  # nosec
        except KeyboardInterrupt:
            logging.info("interrupt!")
            zmqr.stop()
            executor.shutdown()
            x.join()



import argparse
import json
import logging
import os
import queue
import socket
import struct
import sys
import threading
import time

import bjoern
import falcon
from falcon_cors import CORS

from gamutrf.__init__ import __version__
from gamutrf.birdseye_rssi import BirdsEyeRSSI
from gamutrf.birdseye_rssi import FLOAT_SIZE
from gamutrf.birdseye_rssi import MAX_RSSI
from gamutrf.birdseye_rssi import RSSI_UDP_ADDR
from gamutrf.birdseye_rssi import RSSI_UDP_PORT
from gamutrf.mqtt_reporter import MQTTReporter
from gamutrf.sdr_recorder import get_recorder
from gamutrf.sdr_recorder import RECORDER_MAP

WORKER_NAME = os.getenv("WORKER_NAME", socket.gethostbyname(socket.gethostname()))
ORCHESTRATOR = os.getenv("ORCHESTRATOR", "orchestrator")
ANTENNA = os.getenv("ANTENNA", "")


def argument_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--loglevel",
        "-l",
        help="Set logging level",
        choices=["critical", "error", "warning", "info", "debug"],
        default="info",
    )
    parser.add_argument(
        "--antenna", "-a", help="Antenna make/model", type=str, default=ANTENNA
    )
    parser.add_argument(
        "--name", "-n", help="Name for the worker", type=str, default=WORKER_NAME
    )
    parser.add_argument(
        "--path",
        "-P",
        help="Path prefix for writing out samples to",
        type=str,
        default="/data/gamutrf",
    )
    parser.add_argument(
        "--port", "-p", help="Port to run the API webserver on", type=int, default=8000
    )
    parser.add_argument(
        "--rotate_secs",
        help="If > 0, rotate storage directories every N seconds",
        type=int,
        default=3600,
    )
    parser.add_argument(
        "--sdr",
        "-s",
        help=f"Specify SDR to record with {list(RECORDER_MAP.keys())} or file",
        type=str,
        default="ettus",
    )
    parser.add_argument(
        "--sdrargs",
        help=f"optional SDR arguments",
        type=str,
        default="",
    )
    parser.add_argument(
        "--freq_excluded",
        "-e",
        help='Freq range to exclude in MHz (e.g. "100-200")',
        action="append",
        default=[],
    )
    parser.add_argument("--gain", "-g", help="Gain in dB", default=30, type=int)
    parser.add_argument(
        "--mean_window", "-m", help="birdseye mean window size", default=128, type=int
    )
    parser.add_argument(
        "--rxb", help="Receive buffer size", default=int(1024 * 1024 * 10), type=int
    )
    parser.add_argument(
        "--qsize", help="Max request queue size", default=int(2), type=int
    )
    parser.add_argument(
        "--mqtt_server",
        help="MQTT server to report RSSI",
        default=ORCHESTRATOR,
        type=str,
    )
    parser.add_argument(
        "--gps_server",
        help="GPS Server to get lat,long, and heading",
        default=ORCHESTRATOR,
        type=str,
    )
    parser.add_argument(
        "--rssi_interval",
        help="rate limit in seconds for RSSI updates to MQTT",
        default=1.0,
        type=float,
    )
    parser.add_argument(
        "--rssi_throttle",
        help="rate limit RSSI calculations to 1 in n",
        default=10,
        type=int,
    )
    parser.add_argument(
        "--rssi_threshold", help="RSSI reporting threshold", default=-45, type=float
    )
    external_rssi_parser = parser.add_mutually_exclusive_group(required=False)
    external_rssi_parser.add_argument(
        "--rssi_external",
        dest="rssi_external",
        action="store_true",
        default=True,
        help="proxy external RSSI",
    )
    external_rssi_parser.add_argument(
        "--no-rssi_external",
        dest="rssi_external",
        action="store_false",
        help="do not use proxy external RSSI",
    )
    agc_parser = parser.add_mutually_exclusive_group(required=False)
    agc_parser.add_argument(
        "--agc", dest="agc", action="store_true", default=True, help="use AGC"
    )
    agc_parser.add_argument(
        "--no-agc", dest="agc", action="store_false", help="do not use AGC"
    )
    parser.add_argument(
        "--sigmf",
        dest="sigmf",
        default=True,
        action=argparse.BooleanOptionalAction,
        help="add sigmf meta file",
    )
    parser.add_argument(
        "--use_external_gps",
        dest="use_external_gps",
        default=False,
        action=argparse.BooleanOptionalAction,
        help="Use external Pixhawk/MAVLINK GPS",
    )
    parser.add_argument(
        "--use_external_heading",
        dest="use_external_heading",
        default=False,
        action=argparse.BooleanOptionalAction,
        help="Use external (Pixhawk/MAVLINK) heading",
    )
    parser.add_argument(
        "--external_gps_server",
        dest="external_gps_server",
        default=ORCHESTRATOR,
        type=str,
        help="server to query for external GPS data",
    )
    parser.add_argument(
        "--external_gps_server_port",
        dest="external_gps_server_port",
        default="8888",
        type=str,
        help="server port to query for external GPS data",
    )

    return parser


class Endpoints:
    @staticmethod
    def on_get(_req, resp):
        endpoints = []
        for path in API.paths():
            endpoints.append(API.version() + path)

        resp.text = json.dumps(endpoints)
        resp.content_type = falcon.MEDIA_TEXT
        resp.status = falcon.HTTP_200


class Info:
    def __init__(self, arguments):
        self.arguments = arguments

    def on_get(self, _req, resp):
        resp.text = json.dumps(
            {
                "version": __version__,
                "sdr": self.arguments.sdr,
                "path_prefix": self.arguments.path,
                "freq_excluded": self.arguments.freq_excluded,
            }
        )
        resp.content_type = falcon.MEDIA_TEXT
        resp.status = falcon.HTTP_200


class Action:
    def __init__(self, arguments, q, sdr_recorder):
        self.arguments = arguments
        self.q = q
        self.sdr_recorder = sdr_recorder
        self.action = None

    def on_get(self, _req, resp, center_freq, sample_count, sample_rate):
        # TODO check if chosen SDR can do the supplied sample_count
        resp.content_type = falcon.MEDIA_JSON
        resp.status = falcon.HTTP_400

        status = None
        if self.q.full():
            status = "Request queue is full"
        else:
            status = self.sdr_recorder.validate_request(
                self.arguments.freq_excluded, center_freq, sample_count, sample_rate
            )

        if status is None:
            self.q.put(
                {
                    "action": self.action,
                    "center_freq": int(center_freq),
                    "sample_count": int(sample_count),
                    "sample_rate": int(sample_rate),
                }
            )
            status = "Requsted recording"
            resp.status = falcon.HTTP_200

        resp.text = json.dumps({"status": status})


class Record(Action):
    def __init__(self, arguments, q, sdr_recorder):
        super().__init__(arguments, q, sdr_recorder)
        self.action = "record"


class Rssi(Action):
    def __init__(self, arguments, q, sdr_recorder):
        super().__init__(arguments, q, sdr_recorder)
        self.action = "record"


class API:
    def __init__(self, arguments):
        self.arguments = arguments
        self.mqtt_reporter = MQTTReporter(
            name=self.arguments.name,
            mqtt_server=self.arguments.mqtt_server,
            gps_server=self.arguments.gps_server,
            compass=True,
            use_external_gps=self.arguments.use_external_gps,
            use_external_heading=self.arguments.use_external_heading,
            external_gps_server=self.arguments.external_gps_server,
            external_gps_server_port=self.arguments.external_gps_server_port,
        )
        self.q = queue.Queue(self.arguments.qsize)
        self.sdr_recorder = get_recorder(
            self.arguments.sdr, self.arguments.sdrargs, self.arguments.rotate_secs
        )
        self.start_time = time.time()
        cors = CORS(allow_all_origins=True)
        self.app = falcon.App(middleware=[cors.middleware])
        routes = self.routes()
        for route, handler in routes.items():
            self.app.add_route(self.version() + route, handler)

    def run_recorder(self, record_func):
        logging.info("run recorder")
        while True:
            logging.info("awaiting request")
            action_args = self.q.get()
            action = action_args["action"]
            if action == "record":
                self.serve_recording(record_func, action_args)
            elif action == "rssi":
                self.serve_rssi(action_args)
            else:
                logging.error("no such action: %s", action)

    def record(self, center_freq, sample_count, sample_rate=20e6):
        return self.sdr_recorder.run_recording(
            self.arguments.path,
            sample_rate,
            sample_count,
            center_freq,
            self.arguments.gain,
            self.arguments.agc,
            self.arguments.rxb,
            self.arguments.sigmf,
            self.arguments.sdr,
            self.arguments.antenna,
        )

    def serve_recording(self, record_func, record_args):
        logging.info(f"got a request: {record_args}")
        record_status = record_func(**record_args)
        if record_status == -1:
            # TODO this only kills the thread, not the main process
            return
        record_args.update(vars(self.arguments))
        self.mqtt_reporter.publish("gamutrf/record", record_args)
        self.mqtt_reporter.log(
            self.arguments.path, "record", self.start_time, record_args
        )

    def report_rssi(self, record_args, reported_rssi, reported_time):
        logging.info(f'reporting RSSI {reported_rssi} for {record_args["center_freq"]}')
        record_args.update({"rssi": reported_rssi, "time": reported_time})
        record_args.update(vars(self.arguments))
        self.mqtt_reporter.publish("gamutrf/rssi", record_args)
        self.mqtt_reporter.log(
            self.arguments.path, "rssi", self.start_time, record_args
        )

    def process_rssi(self, record_args, sock):
        last_rssi_time = 0
        duration = 0
        if record_args["sample_count"]:
            duration = float(record_args["sample_count"]) / float(
                record_args["sample_rate"]
            )
        start_time = time.time()
        while self.q.empty():
            rssi_raw, _ = sock.recvfrom(FLOAT_SIZE)
            rssi = struct.unpack("f", rssi_raw)[0]
            if rssi < self.arguments.rssi_threshold:
                continue
            if rssi > MAX_RSSI:
                continue
            now = time.time()
            if duration and now - start_time > duration:
                break
            now_diff = now - last_rssi_time
            if now_diff < self.arguments.rssi_interval:
                continue
            self.report_rssi(record_args, rssi, now)
            last_rssi_time = now

    def proxy_rssi(self, rssi_addr, record_args):
        with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as sock:
            # codeql[py/bind-socket-all-network-interfaces]
            sock.bind((rssi_addr, RSSI_UDP_PORT))  # nosec
            self.process_rssi(record_args, sock)

    def serve_rssi(self, record_args):
        logging.info(f"got request {record_args}")
        if self.arguments.rssi_external:
            logging.info("proxying external RSSI")
            # codeql[py/bind-socket-all-network-interfaces]
            self.proxy_rssi("0.0.0.0", record_args)  # nosec
        else:
            center_freq = int(record_args["center_freq"])
            try:
                rssi_server = BirdsEyeRSSI(
                    self.arguments,
                    record_args["sample_rate"],
                    center_freq,
                    agc=self.arguments.agc,
                    rssi_throttle=self.arguments.rssi_throttle,
                )
            except RuntimeError as err:
                logging.error("could not initialize RSSI server: %s", err)
                return
            rssi_server.start()
            logging.info(
                f"serving RSSI for {center_freq}Hz over threshold {self.arguments.rssi_threshold} with AGC {self.arguments.agc}"
            )
            self.proxy_rssi(RSSI_UDP_ADDR, record_args)
            logging.info("RSSI stream stopped")
            rssi_server.stop()
            rssi_server.wait()

    @staticmethod
    def paths():
        return [
            "",
            "/info",
            "/record/{center_freq}/{sample_count}/{sample_rate}",
            "/rssi/{center_freq}/{sample_count}/{sample_rate}",
        ]

    @staticmethod
    def version():
        return "/v1"

    def routes(self):
        p = self.paths()
        endpoints = Endpoints()
        info = Info(self.arguments)
        record = Record(self.arguments, self.q, self.sdr_recorder)
        rssi = Rssi(self.arguments, self.q, self.sdr_recorder)
        funcs = [endpoints, info, record, rssi]
        return dict(zip(p, funcs))

    def run(self):
        logging.info("starting recorder thread")
        recorder_thread = threading.Thread(
            target=self.run_recorder, args=(self.record,)
        )
        recorder_thread.start()

        logging.info("starting API thread")
        bjoern.run(self.app, "0.0.0.0", self.arguments.port)  # nosec
        recorder_thread.join()


def main():
    arguments = argument_parser().parse_args()
    level_int = {"CRITICAL": 50, "ERROR": 40, "WARNING": 30, "INFO": 20, "DEBUG": 10}
    level = level_int.get(arguments.loglevel.upper(), 0)
    logging.basicConfig(level=level, format="%(asctime)s %(message)s")
    try:
        app = API(arguments)
    except ValueError:
        sys.exit(1)
    app.run()



import argparse
import datetime
import os
import re
import numpy as np
from scipy import signal
import matplotlib.pyplot as plt
from gamutrf.utils import is_fft
from gamutrf.sample_reader import get_reader, parse_filename


def read_samples(filename, sample_dtype, sample_bytes, seek_bytes=0, nfft=None, n=None):
    reader = get_reader(filename)

    # with reader(filename) as infile:
    #     infile.seek(int(seek_bytes))
    #     sample_buffer_all = infile.readall()

    with reader(filename) as infile:
        infile.seek(int(seek_bytes))
        # print(f"{n=}")
        # print(f"{nfft=}")
        # print(f"{sample_bytes=}")
        sample_buffer = infile.read(n * nfft * sample_bytes)

        # print(f"{len(sample_buffer)=}, {len(sample_buffer_all)=}")
        # print(len(sample_buffer) == len(sample_buffer_all))
        buffered_samples = int(len(sample_buffer) / sample_bytes)

        if buffered_samples == 0:
            print(filename)
            return None
        if buffered_samples / nfft != n:
            print("incomplete")
            # return None

        x1d = np.frombuffer(sample_buffer, dtype=sample_dtype, count=buffered_samples)
        return x1d["i"] + np.csingle(1j) * x1d["q"]


def argument_parser():
    parser = argparse.ArgumentParser(description="Waterfall plotter from scan data")
    parser.add_argument(
        "--sample_dir", type=str, help="Directory with sample zst files."
    )
    parser.add_argument(
        "--min_freq", type=float, help="Minimum frequency for plotting."
    )
    parser.add_argument(
        "--max_freq", type=float, help="Maximum frequency for plotting."
    )
    parser.add_argument(
        "--sampling_rate", default=100e6, type=float, help="Sampling rate."
    )
    parser.add_argument("--nfft", default=256, type=int, help="FFT length.")
    parser.add_argument(
        "--write_samples",
        default=2048,
        type=int,
        help="Number of samples written during scan.",
    )

    return parser


def main():
    # ARG PARSE PARAMETERS
    parser = argument_parser()
    args = parser.parse_args()
    sample_dir = args.sample_dir
    min_freq = args.min_freq
    max_freq = args.max_freq
    nfft = args.nfft
    n = args.write_samples
    sps = args.sampling_rate

    noverlap = nfft // 8
    db_min = -220
    db_max = -70

    cmap = plt.get_cmap("turbo")

    plot_min_freq = min_freq  # freq - (sps/2)
    plot_max_freq = max_freq  # freq + (sps/2)

    freq_resolution = sps / nfft
    max_idx = round((max_freq - min_freq) / freq_resolution)
    total_time = (nfft * n) / sps
    expected_time_bins = int((nfft * n) / (nfft - noverlap))
    X, Y = np.meshgrid(
        np.linspace(
            plot_min_freq,
            plot_max_freq,
            int((max_freq - min_freq) / freq_resolution + 1),
        ),
        np.linspace(0, total_time, expected_time_bins),
    )
    freq_bin_vals = X[0]
    # print(f"{freq_bin_vals.shape=}{freq_bin_vals=}")

    fig = plt.figure(figsize=(28, 10), dpi=100)
    ax = fig.add_subplot(1, 1, 1)
    box = ax.get_position()
    ax.set_position([box.x0, box.y0, box.width * 0.95, box.height])
    title = ax.text(0.5, 1.02, "", transform=ax.transAxes, va="center", ha="center")
    sm = plt.cm.ScalarMappable(cmap=cmap)
    sm.set_clim(vmin=db_min, vmax=db_max)
    cbar_ax = fig.add_axes([0.87, 0.250, 0.03, 0.5])
    cbar = fig.colorbar(sm, cax=cbar_ax)
    cbar.set_label("dB", rotation=0)

    spec_data = np.empty(X.shape)
    spec_data.fill(np.nan)
    mesh = ax.pcolormesh(X, Y, spec_data, shading="nearest")

    plt.show(block=False)
    plt.pause(0.1)
    background = fig.canvas.copy_from_bbox(fig.bbox)
    ax.draw_artist(mesh)
    fig.canvas.blit(ax.bbox)

    processed_files = []
    while True:
        sample_files = [
            f
            for f in sorted(os.listdir(sample_dir))
            if f.startswith("sample") and f.endswith(".zst")
        ]

        needs_processing = []
        processing_batch = []
        for f in sample_files:
            f = os.path.join(sample_dir, f)
            # center_frequency, sample_rate, sample_dtype, sample_bytes, sample_type, sample_bits, _, timestamp = parse_filename(f)
            file_info = parse_filename(f)
            center_frequency = file_info["center_frequency"]
            sample_rate = file_info["sample_rate"]
            if (
                (
                    ((center_frequency + (sample_rate / 2)) >= min_freq)
                    and ((center_frequency + (sample_rate / 2)) <= max_freq)
                )
                or (
                    ((center_frequency - (sample_rate / 2)) >= min_freq)
                    and ((center_frequency - (sample_rate / 2)) <= max_freq)
                )
            ) and f not in processed_files:
                if (
                    not processing_batch
                    or processing_batch[-1]["center_frequency"] < center_frequency
                ):
                    processing_batch.append(file_info)
                else:
                    needs_processing.append(processing_batch)
                    processing_batch = [file_info]

        # for n in needs_processing:
        #     print(f"{len(n)=}\n{n=}\n\n")

        # process files
        if needs_processing:
            # f = needs_processing[0][0]["filename"]
            # print(f"{f=}")
            spec_data.fill(np.nan)
            for file_info in needs_processing[0]:
                filename = file_info["filename"]
                sample_dtype = file_info["sample_dtype"]
                sample_bytes = file_info["sample_len"]
                center_frequency = file_info["center_frequency"]
                sample_rate = file_info["sample_rate"]
                sample_type = file_info["sample_type"]
                sample_bits = file_info["sample_bits"]
                timestamp = file_info["timestamp"]
                # filename, center_frequency, sample_rate, sample_dtype, sample_bytes, sample_type, sample_bits, _, timestamp = parse_filename(f)

                samples = read_samples(
                    filename, sample_dtype, sample_bytes, seek_bytes=0, nfft=nfft, n=n
                )
                if samples is None:
                    continue
                freq_bins, t_bins, S = signal.spectrogram(
                    samples,
                    sample_rate,
                    window=signal.hann(int(nfft), sym=True),
                    nperseg=nfft,
                    noverlap=noverlap,
                    detrend="constant",
                    return_onesided=False,
                )
                freq_bins = np.fft.fftshift(freq_bins)
                # print(f"{freq_bins.shape=}{center_frequency + freq_bins=}")

                idx = np.array(
                    [
                        round((item - min_freq) / freq_resolution)
                        for item in freq_bins + center_frequency
                    ]
                ).astype(int)
                # print(f"{idx=}")
                # print(f"{np.where(idx>=0)}")
                # print(f"{(freq_bins+center_frequency)[np.where(idx>=0)]=}")
                # print(f"{freq_bins.shape=}, {S.shape=}")

                S = np.fft.fftshift(S, axes=0)
                S = S.T
                S = 10 * np.log10(S)
                # S_norm = (S-np.min(S))/(np.max(S) - np.min(S))
                print(f"{np.min(S)=}, {np.max(S)=}")
                S_norm = (S - db_min) / (db_max - db_min)
                # plt.figure()
                # plt.pcolormesh(t_bins, freq_bins, S)
                # plt.show()
                # time.sleep(10)
                # print(f"{filename=}, {round((max_freq-min_freq)/freq_resolution)=}")
                # print(f"{np.flatnonzero(idx>=0)=}")
                # print(f"{idx[np.flatnonzero(idx>=0)]=}")
                spec_data[
                    : S_norm.shape[0],
                    idx[np.flatnonzero((idx >= 0) & (idx <= max_idx))],
                ] = S_norm[:, np.flatnonzero((idx >= 0) & (idx <= max_idx))]
                processed_files.append(filename)

            fig.canvas.restore_region(background)
            # spec_data = np.empty(X.shape)

            try:
                mesh.set_array(cmap(spec_data))
            except Exception as e:
                print(f"{e}")
                continue
            # print(f"{filename=}, {idx[np.flatnonzero(idx>=0)]=}\n")
            title.set_text(str(datetime.datetime.fromtimestamp(timestamp)))
            ax.draw_artist(title)
            ax.draw_artist(mesh)
            fig.canvas.blit(ax.title.axes.figure.bbox)
            fig.canvas.blit(fig.bbox)
            fig.canvas.flush_events()
            # time.sleep(3)

            # print(f"{freq_bins=}")
            # print(f"{freq_bins.shape=}")
            # print(f"{t_bins=}")
            # print(f"{t_bins.shape=}")

            # print(f"{samples.shape=}")

            # print(f"{processed_files=}")
        else:
            print("waiting")
            quit()


if __name__ == "__main__":
    main()



import datetime
import json
import logging
import os
import subprocess
import tempfile
import time
from urllib.parse import urlparse

import sigmf
import numpy as np
import matplotlib
import matplotlib.pyplot as plt

from gamutrf.sigwindows import freq_excluded
from gamutrf.sigwindows import parse_freq_excluded
from gamutrf.utils import (
    ETTUS_ANT,
    ETTUS_ARGS,
    MPL_BACKEND,
    WIDTH,
    HEIGHT,
    DPI,
    DS_PIXELS,
    endianstr,
)

IMSHOW_INTERPOLATION = os.getenv("IMSHOW_INTERPOLATION", "bilinear")
NFFT = int(os.getenv("NFFT", "0"))
NFFT_OVERLAP = 512
SAMPLE_TYPE = "i16"
MIN_SAMPLE_RATE = int(1e6)
MAX_SAMPLE_RATE = int(30 * 1e6)
FFT_FILE = "/dev/shm/fft.dat"  # nosec


class SDRRecorder:
    def __init__(self, sdrargs, rotate_secs):
        self.tmpdir = tempfile.TemporaryDirectory()
        self.zst_fifo = os.path.join(self.tmpdir.name, "zstfifo")
        self.sdrargs = sdrargs
        self.rotate_secs = rotate_secs
        os.mkfifo(self.zst_fifo)

    @staticmethod
    def validate_request(freqs_excluded, center_freq, sample_count, sample_rate):
        for arg in [center_freq, sample_count, sample_rate]:
            try:
                int(float(arg))
            except ValueError:
                return "Invalid values in request"
        if freq_excluded(int(center_freq), parse_freq_excluded(freqs_excluded)):
            return "Requested frequency is excluded"
        if int(sample_rate) < MIN_SAMPLE_RATE or int(sample_rate) > MAX_SAMPLE_RATE:
            return "sample rate {sample_rate} out of range {MIN_SAMPLE_RATE} to {MAX_SAMPLE_RATE}"
        duration_sec = int(sample_count) / int(sample_rate)
        if duration_sec < 1:
            return "cannot record for less than 1 second"
        return None

    @staticmethod
    def get_sample_file(path, epoch_time, center_freq, sample_rate, sdr, antenna, gain):
        return os.path.join(
            path,
            f"gamutrf_recording_{sdr}_{antenna}_gain{gain}_{epoch_time}_{int(center_freq)}Hz_{int(sample_rate)}sps.{SAMPLE_TYPE}.zst",
        )

    def record_args(
        self, sample_file, sample_rate, sample_count, center_freq, gain, agc, rxb
    ):
        raise NotImplementedError

    def write_recording(
        self, sample_file, sample_rate, sample_count, center_freq, gain, agc, rxb
    ):
        record_status = -1
        args = self.record_args(
            self.zst_fifo, sample_rate, sample_count, center_freq, gain, agc, rxb
        )
        dotfile = os.path.join(
            os.path.dirname(sample_file), "." + os.path.basename(sample_file)
        )
        zstd_args = ["nice", "zstd", "-1", self.zst_fifo, "-o", dotfile]
        logging.info("starting recording: %s", args)
        with subprocess.Popen(zstd_args, stdin=subprocess.DEVNULL) as zstd_proc:
            record_status = subprocess.check_call(args)
            zstd_proc.communicate()
        if os.path.exists(dotfile):
            os.rename(dotfile, sample_file)
        return record_status

    @staticmethod
    def fft_spectrogram(
        sample_file, fft_file, sample_count, sample_rate, center_freq, nfft
    ):
        def matplotlib_gc(fig, im):
            im.remove()
            fig.clear()
            plt.close()
            plt.cla()
            plt.clf()

        def get_fig(width, height):
            fig = plt.figure()
            fig.set_size_inches(width, height)
            axes = fig.add_subplot(111)
            return (fig, axes)

        def imshow(axes, i, extent):
            return axes.imshow(
                i,
                cmap="turbo",
                origin="lower",
                extent=extent,
                aspect="auto",
                interpolation=IMSHOW_INTERPOLATION,
            )

        def plot_fft(i, sample_file, extent):
            png_file = sample_file.replace(".zst", ".png")
            logging.info("generating spectrogram: %s", png_file)
            fig, axes = get_fig(WIDTH, HEIGHT)
            axes.set_xlabel("time (s)")
            axes.set_ylabel("freq (MHz)")
            axes.axis("auto")
            axes.minorticks_on()
            im = imshow(axes, i, extent)
            plt.colorbar(im, ax=axes)
            plt.sci(im)
            plt.savefig(png_file, dpi=DPI)
            matplotlib_gc(fig, im)

        def plot_ds_fft(i, sample_file, extent):
            ds_png_file = sample_file.replace(".zst", ".ds.png")
            logging.info("generating downsampled spectrogram: %s", ds_png_file)
            fig, axes = get_fig(DS_PIXELS / DPI, DS_PIXELS / DPI)
            axes.set_axis_off()
            axes.get_xaxis().set_visible(False)
            axes.get_yaxis().set_visible(False)
            im = imshow(axes, i, extent)
            plt.sci(im)
            plt.subplots_adjust(left=0, right=1, bottom=0, top=1)
            plt.savefig(ds_png_file, dpi=DPI)
            matplotlib_gc(fig, im)

        if os.path.exists(fft_file):
            matplotlib.use(MPL_BACKEND)
            i = np.memmap(fft_file, dtype=np.float32, mode="r+")
            i = np.roll(i.reshape(-1, nfft).swapaxes(0, 1), int(nfft / 2), 0)
            fc = center_freq / 1e6
            fo = sample_rate / 1e6 / 2
            extent = (0, sample_count / sample_rate, fc - fo, fc + fo)
            plot_fft(i, sample_file, extent)
            plot_ds_fft(i, sample_file, extent)
            os.remove(fft_file)

    def run_recording(
        self,
        path,
        sample_rate,
        sample_count,
        center_freq,
        gain,
        agc,
        rxb,
        sigmf_,
        sdr,
        antenna,
    ):
        epoch_time = int(time.time())
        meta_time = datetime.datetime.utcnow().isoformat() + "Z"
        if self.rotate_secs:
            ts_dir = int(epoch_time / self.rotate_secs) * self.rotate_secs
            path = os.path.join(path, str(ts_dir))
            if not os.path.exists(path):
                os.makedirs(path)
        sample_file = self.get_sample_file(
            path, str(epoch_time), center_freq, sample_rate, sdr, antenna, gain
        )
        record_status = -1
        try:
            record_status = self.write_recording(
                sample_file, sample_rate, sample_count, center_freq, gain, agc, rxb
            )
            if NFFT:
                self.fft_spectrogram(
                    sample_file, FFT_FILE, sample_count, sample_rate, center_freq, NFFT
                )

            if sigmf_:
                if os.path.exists(sample_file):
                    meta = sigmf.SigMFFile(
                        skip_checksum=True,  # expensive for large files
                        data_file=sample_file,
                        global_info={
                            sigmf.SigMFFile.DATATYPE_KEY: "_".join(
                                ("c" + SAMPLE_TYPE, endianstr())
                            ),
                            sigmf.SigMFFile.SAMPLE_RATE_KEY: sample_rate,
                            sigmf.SigMFFile.VERSION_KEY: sigmf.__version__,
                        },
                    )
                    # TODO: add capture_details, source_file and gain when supported.
                    meta.add_capture(
                        0,
                        metadata={
                            sigmf.SigMFFile.FREQUENCY_KEY: center_freq,
                            sigmf.SigMFFile.DATETIME_KEY: meta_time,
                        },
                    )
                    meta.tofile(sample_file + ".sigmf-meta")
                else:
                    logging.error("{sample_file} missing, cannot write sigmf file")
        except subprocess.CalledProcessError as err:
            logging.error("record failed: %s", err)
        logging.info("record status: %d", record_status)
        return (record_status, sample_file)


class EttusRecorder(SDRRecorder):
    def __init__(self, sdrargs, rotate_secs):
        super().__init__(sdrargs, rotate_secs)
        self.worker_subprocess = None
        self.last_worker_line = None
        if not self.sdrargs:
            self.sdrargs = ETTUS_ARGS
        try:
            subprocess.check_call(
                [
                    "/usr/local/bin/uhd_sample_recorder",
                    "--rate",
                    str(1e6),
                    "--args",
                    self.sdrargs,
                    "--ant",
                    ETTUS_ANT,
                    "--duration",
                    "1",
                    "--null",
                    "--fftnull",
                    "--novkfft",
                    "--nfft",
                    "0",
                ]
            )
        except subprocess.CalledProcessError:
            raise ValueError

    def record_args(
        self, sample_file, sample_rate, sample_count, center_freq, gain, _agc, rxb
    ):
        # Ettus "nsamps" API has an internal limit, so translate "stream for druation".
        duration = round(sample_count / sample_rate)
        rxb = min(rxb, sample_rate)

        args = [
            "/usr/local/bin/uhd_sample_recorder",
            "--json",
            "--rate",
            str(sample_rate),
            "--bw",
            str(sample_rate),
            "--gain",
            str(gain),
            "--args",
            self.sdrargs,
            "--ant",
            ETTUS_ANT,
            "--spb",
            str(rxb),
        ]

        json_args = {
            "file": sample_file,
            "duration": duration,
            "freq": center_freq,
        }
        if NFFT:
            json_args.update(
                {
                    "nfft": NFFT,
                    "nfft_overlap": NFFT_OVERLAP,
                    "fft_file": FFT_FILE,
                }
            )
        return (args, json_args)

    def write_recording(
        self, sample_file, sample_rate, sample_count, center_freq, gain, agc, rxb
    ):
        # Ettus doesn't need a wrapper, it can do its own zst compression.
        record_status = -1
        args, json_args = self.record_args(
            sample_file, sample_rate, sample_count, center_freq, gain, agc, rxb
        )

        try:

            def worker_read():
                self.last_worker_line = (
                    self.worker_subprocess.stdout.readline().decode("utf-8").strip()
                )
                logging.info(self.last_worker_line)

            def worker_write(s):
                self.worker_subprocess.stdin.write(("%s\n" % s).encode("utf-8"))
                self.worker_subprocess.stdin.flush()

            if self.worker_subprocess is None:
                logging.info("starting worker subprocess: %s", args)
                self.worker_subprocess = subprocess.Popen(
                    args, stdin=subprocess.PIPE, stdout=subprocess.PIPE
                )
                worker_read()

            logging.info("starting recording: %s", json_args)
            worker_write(json.dumps(json_args))
            worker_read()
            last_error = json.loads(self.last_worker_line).get("last_error")
            if not last_error:
                record_status = 0
        except (
            subprocess.SubprocessError,
            BrokenPipeError,
            json.decoder.JSONDecodeError,
        ) as e:
            logging.error(e)
            if self.worker_subprocess:
                self.worker_subprocess.kill()
                self.worker_subprocess = None
        return record_status


class BladeRecorder(SDRRecorder):
    def record_args(
        self, sample_file, sample_rate, sample_count, center_freq, gain, agc, _rxb
    ):
        gain_args = [
            "-e",
            "set agc rx off",
            "-e",
            f"set gain rx {gain}",
        ]
        if agc:
            gain_args = [
                "-e",
                "set agc rx on",
            ]
        return (
            ["bladeRF-cli"]
            + gain_args
            + [
                "-e",
                f"set samplerate rx {sample_rate}",
                "-e",
                f"set bandwidth rx {sample_rate}",
                "-e",
                f"set frequency rx {center_freq}",
                "-e",
                f"rx config file={sample_file} format=bin n={sample_count}",
                "-e",
                "rx start",
                "-e",
                "rx wait",
            ]
        )


class LimeRecorder(SDRRecorder):
    def record_args(
        self, sample_file, sample_rate, sample_count, center_freq, gain, agc, _rxb
    ):
        gain_args = []
        if gain:
            gain_args = [
                "-g",
                f"{gain}",
            ]
        return (
            ["/usr/local/bin/LimeStream"]
            + gain_args
            + [
                "-f",
                f"{center_freq}",
                "-s",
                f"{sample_rate}",
                "-C",
                f"{sample_count}",
                "-r",
                f"{sample_file}",
            ]
        )


class FileTestRecorder(SDRRecorder):
    def record_args(
        self, sample_file, sample_rate, sample_count, center_freq, gain, agc, rxb
    ):
        args = [
            "dd",
            f"if={urlparse(self.sdrargs).path}",
            f"of={sample_file}",
            f"count={int(sample_count)}",
            f"bs={int(sample_rate)}",
        ]
        return args


RECORDER_MAP = {
    "ettus": EttusRecorder,
    "bladerf": BladeRecorder,
    "lime": LimeRecorder,
}


def get_recorder(recorder_name, sdrargs=None, rotate_secs=0):
    try:
        return RECORDER_MAP[recorder_name](sdrargs, rotate_secs)
    except KeyError:
        return FileTestRecorder(recorder_name, rotate_secs)
